package ca.mcgill.sel.aspectoptimaomtt.interfacing;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.FutureTask;

import ca.mcgill.sel.aspectoptima.aspects.context.Context;
import ca.mcgill.sel.aspectoptima.aspects.context.Participant;
import ca.mcgill.sel.aspectoptima.aspects.outcomeaware.BinaryOutcomeImpl;
import ca.mcgill.sel.aspectoptima.aspects.outcomeaware.Outcome;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactionParticipant;
import ca.mcgill.sel.aspectoptima.interfacing.AspectOPTIMA;
import ca.mcgill.sel.aspectoptima.interfacing.AspectOPTIMAInternal;
import ca.mcgill.sel.aspectoptima.interfacing.ParticipantPool;
import ca.mcgill.sel.aspectoptimaomtt.aspects.collaborative.CollaborativeContext;
import ca.mcgill.sel.aspectoptimaomtt.aspects.collaborative.CollaborativeParticipant;
import ca.mcgill.sel.aspectoptimaomtt.aspects.openmultithreadedtransaction.OMTTContext;
import ca.mcgill.sel.aspectoptimaomtt.aspects.openmultithreadedtransaction.OMTTParticipant;
import ca.mcgill.sel.aspectoptimaomtt.aspects.outcomevotable.OutcomeVotableContext;
import ca.mcgill.sel.aspectoptimaomtt.aspects.outcomevotable.OutcomeVotableParticipant;
import ca.mcgill.sel.aspectoptimaomtt.aspects.spawnsupporting.SpawnSupportingContext;
import ca.mcgill.sel.aspectoptimaomtt.aspects.spawnsupporting.SpawnSupportingParticipant;

public class AspectOPTIMAOMTT {
	static Map<Object, OMTTThread> keyToThreadMap = new HashMap<Object, OMTTThread>();
	
	/**
	 * Runs the passed runnable in a new OMTTThread. <br>
	 * This method is the entry point to the OMTT framework and <b>every code that uses OMTT features must have been passed to his method</b>
	 *  in a runnable or it must have been executed by the framework itself through spawned threads.
	 * @param runnable
	 * @return true if the code was successfully run in a new OMTTThread, false in case of any error
	 */
	public static boolean run(Runnable runnable) {
		try {
			return createAndStartNewThreadWithoutResult(runnable) != null;
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Begins an OpenMultithreadedTransaction using the passed update strategy and concurrency control mechanism <br>
	 * and sets the minimal and maximal number of participants to the provided values.
	 * Allows for nested transactions if the parameter nested is true.
	 * @param updateStrategy "CHECKPOINTING" or "DEFERRING"
	 * @param concurrencyControl "TWOPHASELOCKING" or "OPTIMISTICVALIDATION"
	 * @param nested whether nested child transactions should be allowed or not
	 * @return true if the corresponding transaction was successfully started, false in case of any error
	 */
	public static boolean beginTransactionInCurrentThread(String updateStrategy, String concurrencyControl, boolean nested, int minParticipantCount, int maxParticipantCount) {
		try {
			OMTTParticipant participant = AspectOPTIMAOMTTInternal.getOMTTParticipant(updateStrategy, concurrencyControl, nested);
			/* content of invisible createContext */
			String participantClassName = participant.getClass().getName();
			String contextImplClassName = participantClassName.replaceFirst("Participant", "Context");
			Class<? extends OMTTContext> implementationClass = (Class<? extends OMTTContext> ) Class.forName(contextImplClassName);
			Constructor<? extends OMTTContext> constructor = implementationClass.getDeclaredConstructor((Class<?>[]) null);
			constructor.setAccessible(true);
			OMTTContext context = constructor.newInstance((Object[]) null);
			context.setMinParticipantCount(minParticipantCount);
			context.setMaxParticipantCount(maxParticipantCount);
			/* enter context */
			participant.enterContext(context);
			return true;
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Begins an OpenMultithreadedTransaction using the passed update strategy and concurrency control mechanism. <br>
	 * Allows for nested transactions if the parameter nested is true.
	 * @param updateStrategy "CHECKPOINTING" or "DEFERRING"
	 * @param concurrencyControl "TWOPHASELOCKING" or "OPTIMISTICVALIDATION"
	 * @param nested whether nested child transactions should be allowed or not
	 * @return true if the corresponding transaction was successfully started, false in case of any error
	 */
	public static boolean beginTransactionInCurrentThread(String updateStrategy, String concurrencyControl, boolean nested) {
		try {
			OMTTParticipant participant = AspectOPTIMAOMTTInternal.getOMTTParticipant(updateStrategy, concurrencyControl, nested);
			participant.beginTransaction();
			return true;
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Begins an OpenMultithreadedTransaction in a new thread using the passed update strategy and concurrency control mechanism
	 * and sets the minimal and maximal number of participants to the provided values.
	 * The returned new OMTTThread provides methods for enqueueing new jobs. <br>
	 * Allows for nested transactions if the parameter nested is true.
	 * @param updateStrategy "CHECKPOINTING" or "DEFERRING"
	 * @param concurrencyControl "TWOPHASELOCKING" or "OPTIMISTICVALIDATION"
	 * @param nested whether nested child transactions should be allowed or not
	 * @return the new OMTTThread (extends Thread) in which the transaction was started if it was started successfully, null in case of any error
	 */
	public static OMTTThread beginTransactionInNewThread(final String updateStrategy, final String concurrencyControl, final boolean nested, final int minParticipantCount, final int maxParticipantCount) {
		try {
			Runnable beginTransactionRunnable =
				new Runnable() { 
					public void run() {
						beginTransactionInCurrentThread(updateStrategy, concurrencyControl, nested, minParticipantCount, maxParticipantCount);
					}
				};
			return createAndStartNewThreadWithoutResult(beginTransactionRunnable);
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Begins an OpenMultithreadedTransaction in a new thread using the passed update strategy and concurrency control mechanism.
	 * The returned new OMTTThread provides methods for enqueueing new jobs. <br>
	 * Allows for nested transactions if the parameter nested is true.
	 * @param updateStrategy "CHECKPOINTING" or "DEFERRING"
	 * @param concurrencyControl "TWOPHASELOCKING" or "OPTIMISTICVALIDATION"
	 * @param nested whether nested child transactions should be allowed or not
	 * @return the new OMTTThread (extends Thread) in which the transaction was started if it was started successfully, null in case of any error
	 */
	public static OMTTThread beginTransactionInNewThread(final String updateStrategy, final String concurrencyControl, final boolean nested) {
		try {
			Future<Boolean> beginTransactionFuture = 
				new FutureTask<Boolean>(
					new Callable<Boolean>() { 
						public Boolean call() {
							return beginTransactionInCurrentThread(updateStrategy, concurrencyControl, nested);
						}
					}
				);
			return createAndStartNewThread(beginTransactionFuture);
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return null;
	}
	
	private static OMTTThread createAndStartNewThread(Future<Boolean> future) throws InterruptedException, ExecutionException {
		OMTTThread newthread = new OMTTThread((Runnable) future);
		newthread.start();
		boolean success = future.get();
		return success ? newthread : null;
	}
	
	static OMTTThread createAndStartNewThreadWithoutResult(Runnable runnable) throws InterruptedException, ExecutionException {
		OMTTThread newthread = new OMTTThread(runnable);
		newthread.start();
		return newthread;
	}
	
	/**
	 * Begins a nested OpenMultithreadedTransaction as a child of the OpenMultithreadedTransaction that is 
	 * associated to the currently active thread using the configuration parameters of the parent transaction.
	 * @return true if the corresponding transaction was successfully started, false in case of any error
	 */
	public static boolean beginChildTransaction() {
		return AspectOPTIMA.beginChildTransaction();
	}
	
	/**
	 * Lets the currently executing thread join the OpenMultithreadedTransaction of the thread that was passed as an argument.
	 * @param threadToJoin the thread that is associated to the OpenMultithreadedTransaction that should be joined
	 * @return true if the corresponding transaction was successfully joined, false in case of any error
	 */
	public static boolean joinTransactionWithCurrentThread(OMTTThread threadToJoin) {
		try {
			Context context = getContextFromThread(threadToJoin);
			return joinContext(context);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}

	private static Context getContextFromThread(OMTTThread thread) throws InterruptedException, ExecutionException {
		Future<Context> getContextFuture = 
			new FutureTask<Context>(
				new Callable<Context>() {
					public Context call() {
						OMTTParticipant participant = (OMTTParticipant) ParticipantPool.claimParticipant();
						mapKeyToThread(participant, (OMTTThread) Thread.currentThread());
						return participant.getContext();
					}
				}
			);
		thread.enqueue((Runnable) getContextFuture);
		return getContextFuture.get();
	}
	
	private static boolean joinContext(Context context) {
		OMTTParticipant participant = (OMTTParticipant) ParticipantPool.claimParticipant();
		mapKeyToThread(participant, (OMTTThread) Thread.currentThread());
		participant.joinContext((CollaborativeContext) context);
		Context newContext = participant.getContext();
		return (newContext == context) ? true : false;
	}

	/**
	 * Lets the thread that is passed as joiningThread join the OpenMultithreadedTransaction of the thread that was passed as threadToJoin.
	 * @param threadToJoin the thread that is associated to the OpenMultithreadedTransaction that should be joined
	 * @param joiningThread the thread that should join the OpenMultithreadedTransaction of threadToJoin
	 * @return true if the corresponding transaction was successfully joined, false in case of any error
	 */
	public static boolean joinTransactionWithExistingThread(OMTTThread threadToJoin, OMTTThread joiningThread) {
		try {
			final Context context = getContextFromThread(threadToJoin);
			Future<Boolean> joinTransactionFuture = 
				new FutureTask<Boolean>(
					new Callable<Boolean>() { 
						public Boolean call() {
							return joinContext(context);
						}
					}
				);
			joiningThread.enqueue((Runnable) joinTransactionFuture);
			boolean success = joinTransactionFuture.get();
			if (success) {
				return true;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Lets the thread that is passed as joiningThread join the OpenMultithreadedTransaction of the currently executing thread.
	 * @param joiningThread the thread that should join the OpenMultithreadedTransaction of threadToJoin
	 * @return true if the corresponding transaction was successfully joined, false in case of any error
	 */
	public static boolean joinCurrentTransactionWithExistingThread(OMTTThread joiningThread) {
		try {
			OMTTParticipant participant = (OMTTParticipant) ParticipantPool.claimParticipant();
			mapKeyToThread(participant, (OMTTThread) Thread.currentThread());
			final Context context = participant.getContext();
			Future<Boolean> joinTransactionFuture = 
				new FutureTask<Boolean>(
					new Callable<Boolean>() { 
						public Boolean call() {
							return joinContext(context);
						}
					}
				);
			joiningThread.enqueue((Runnable) joinTransactionFuture);
			boolean success = joinTransactionFuture.get();
			if (success) {
				return true;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Spawns a participant of an OpenMultithreadedTransaction using the transaction that is associated to the currently executing thread.
	 * @return the new OMTTThread (extends Thread) that is associated to the spawned participant if it was started successfully, null in case of any error
	 */
	public static OMTTThread spawnTransactionParticipantFromCurrentThread() {
		try {
			return spawnParticipant();
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return null;
	}
	
	private static OMTTThread spawnParticipant() throws InterruptedException, ExecutionException {
		OMTTParticipant currentParticipant = (OMTTParticipant) ParticipantPool.claimParticipant();
		mapKeyToThread(currentParticipant, (OMTTThread) Thread.currentThread());
		final SpawnSupportingParticipant spawnedParticipant = currentParticipant.spawnParticipant();
		Future<Boolean> assignSpawnedPToThreadFuture = 
			new FutureTask<Boolean>(
				new Callable<Boolean>() { 
					public Boolean call() {
						ParticipantPool.setThreadLocalParticipant((Participant) spawnedParticipant);
						mapKeyToThread(spawnedParticipant, (OMTTThread) Thread.currentThread());
						return true;
					}
				}
			);
		return createAndStartNewThread(assignSpawnedPToThreadFuture);
	}
	
	/**
	 * Spawns a participant of an OpenMultithreadedTransaction using the transaction that is associated to the thread that was passed as an argument.
	 * @return the new OMTTThread (extends Thread) that is associated to the spawned participant if it was started successfully, null in case of any error
	 */
	public static OMTTThread spawnAndReturnTransactionParticipantFromExistingThread(OMTTThread spawningThread) {
		try {
			Future<OMTTThread> spawnParticipantFuture = 
				new FutureTask<OMTTThread>(
					new Callable<OMTTThread>() { 
						public OMTTThread call() {
							try {
								return spawnParticipant();
							} catch (Exception e) {
								// swallow
								e.printStackTrace();
							}
							return null;
						}
					}
				);
			spawningThread.enqueue((Runnable) spawnParticipantFuture);
			return spawnParticipantFuture.get();
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return null;
	}
	
	/**
	 * Begins an OpenMultithreadedTransaction in a new thread using the passed update strategy and concurrency control mechanism.
	 * Runs the passed Runnable immediately afterwards in the newly created thread.
	 * This returned new OMTTThread provides methods for enqueueing new jobs. <br>
	 * Allows for nested transactions if the parameter nested is true. <br>
	 * Note that the return value can simply be ignored if the new thread should never execute anything else than the passed Runnable.
	 * @param updateStrategy "CHECKPOINTING" or "DEFERRING"
	 * @param concurrencyControl "TWOPHASELOCKING" or "OPTIMISTICVALIDATION"
	 * @param nested whether nested child transactions should be allowed or not
	 * @return the new OMTTThread (extends Thread) in which the transaction was started if it was started successfully, null in case of any error
	 */
	public static OMTTThread beginAndRunTransactionInNewThread(Runnable r, String updateStrategy, String concurrencyControl, boolean nested) {
		OMTTThread newThread = beginTransactionInNewThread(updateStrategy, concurrencyControl, nested);
		newThread.enqueue(r);
		return newThread;
	}
	
	/**
	 * Begins an OpenMultithreadedTransaction in a new thread using the passed update strategy and concurrency control mechanism
	 * and sets the minimal and maximal number of participants to the provided values.
	 * Runs the passed Runnable immediately afterwards in the newly created thread.
	 * This returned new OMTTThread provides methods for enqueueing new jobs. <br>
	 * Allows for nested transactions if the parameter nested is true. <br>
	 * Note that the return value can simply be ignored if the new thread should never execute anything else than the passed Runnable.
	 * @param updateStrategy "CHECKPOINTING" or "DEFERRING"
	 * @param concurrencyControl "TWOPHASELOCKING" or "OPTIMISTICVALIDATION"
	 * @param nested whether nested child transactions should be allowed or not
	 * @return the new OMTTThread (extends Thread) in which the transaction was started if it was started successfully, null in case of any error
	 */
	public static OMTTThread beginAndRunTransactionInNewThread(Runnable r, String updateStrategy, String concurrencyControl, boolean nested, int minParticipantCount, int maxParticipantCount) {
		OMTTThread newThread = beginTransactionInNewThread(updateStrategy, concurrencyControl, nested, minParticipantCount, maxParticipantCount);
		newThread.enqueue(r);
		return newThread;
	}
	
	/**
	 * Lets the thread that is passed as joiningThread join the OpenMultithreadedTransaction of the thread that was passed as threadToJoin. <br>
	 * Runs the passed Runnable immediately afterwards in the joining thread.
	 * @param threadToJoin the thread that is associated to the OpenMultithreadedTransaction that should be joined
	 * @param joiningThread the thread that should join the OpenMultithreadedTransaction of threadToJoin
	 * @return true if the corresponding transaction was successfully joined, false in case of any error
	 */
	public static boolean joinAndRunTransactionWithExistingThread(Runnable r, OMTTThread threadToJoin, OMTTThread joiningThread) {
		boolean success = joinTransactionWithExistingThread(threadToJoin, joiningThread);
		joiningThread.enqueue(r);
		return success;
	}
	
	/**
	 * Spawns a participant of an OpenMultithreadedTransaction using the transaction that is associated to the currently executing thread.
	 * Runs the passed Runnable in the newly spawned thread.
	 * @return the new OMTTThread (extends Thread) that is associated to the spawned participant if it was started successfully, null in case of any error
	 */
	public static OMTTThread spawnAndRunTransactionParticipantFromCurrentThread(Runnable r) {
		OMTTThread newThread = spawnTransactionParticipantFromCurrentThread();
		newThread.enqueue(r);
		return newThread;
	}
	
	/**
	 * Spawns a participant of an OpenMultithreadedTransaction using the transaction that is associated to the thread that was passed as an argument.
	 * Runs the passed Runnable in the newly spawned thread.
	 * @return the new OMTTThread (extends Thread) that is associated to the spawned participant if it was started successfully, null in case of any error
	 */
	public static OMTTThread spawnRunAndReturnTransactionParticipantFromExistingThread(Runnable r, OMTTThread spawningThread) {
		OMTTThread newThread = spawnAndReturnTransactionParticipantFromExistingThread(spawningThread);
		newThread.enqueue(r);
		return newThread;
	}
	
	/**
	 * Spawns a participant of an OpenMultithreadedTransaction using the transaction that is associated to the thread that was passed as an argument.
	 * Runs the passed Runnable in the newly spawned thread but doesn't wait for the newly spawned thread to return it.
	 */
	public static void spawnRunAndForgetTransactionParticipantFromExistingThread(final Runnable r, OMTTThread spawningThread) {
		try {
			Runnable spawnParticipantRunnable = 
				new Runnable() { 
					public void run() {
						try {
							spawnParticipant();
							r.run();
						} catch (Exception e) {
							// swallow
							e.printStackTrace();
						}
					}
				};
			spawningThread.enqueue(spawnParticipantRunnable);
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
	}
	
	/**
	 * Commits the OpenMultithreadedTransaction that is associated to the currently active thread.
	 * @return true if the current transaction was successfully committed, false in case of any error
	 */
	public static boolean commitTransactionOfCurrentThread() {
		return AspectOPTIMA.commitTransaction();
	}
	
	/**
	 * Commits the OpenMultithreadedTransaction that is associated to the currently active thread.
	 * @return true if the current transaction was successfully committed, false in case of any error
	 */
	public static boolean commitTransactionOfCurrentAndSpawnedThreads() {
		SpawnSupportingParticipant participant = (SpawnSupportingParticipant) ParticipantPool.claimParticipant();
		SpawnSupportingContext context = (SpawnSupportingContext) participant.getContext();
		Collection<CollaborativeParticipant> participants = context.getParticipants();
		ArrayList<Participant> participantsCopy = new ArrayList<Participant>(participants);
		Outcome positiveVote = new BinaryOutcomeImpl(true);
		for (Participant p : participantsCopy) {
			if (participants.contains(p)) {
				if (((SpawnSupportingParticipant) p).isSpawned()) {
					((OutcomeVotableContext) context).registerVote(positiveVote, (OutcomeVotableParticipant) p);
				}
			}
		}
		return AspectOPTIMA.commitTransaction();
	}
	
	
	/**
	 * Commits the OpenMultithreadedTransaction that is associated to the thread that is passed as an argument
	 * @return true if the associated transaction was successfully committed, false in case of any error
	 */
	public static boolean commitTransactionOfForeignThread(OMTTThread thread) {
		try {
			OMTTParticipant participant = getParticipantFromThread(thread);
			participant.commitTransaction();
			return true;
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return false;
	}
	
	private static OMTTParticipant getParticipantFromThread(OMTTThread thread) throws InterruptedException, ExecutionException {
		Future<OMTTParticipant> getParticipantFuture = 
			new FutureTask<OMTTParticipant>(
				new Callable<OMTTParticipant>() {
					public OMTTParticipant call() {
						OMTTParticipant participant = (OMTTParticipant) ParticipantPool.claimParticipant();
						mapKeyToThread(participant, (OMTTThread) Thread.currentThread());
						return participant;
					}
				}
			);
		thread.enqueue((Runnable) getParticipantFuture);
		return getParticipantFuture.get();
	}
	
	/**
	 * Aborts the transaction that is associated to the currently active thread.
	 * @return true if the current transaction was successfully aborted, false in case of any error
	 */
	public static boolean abortTransactionOfCurrentThread() {
		return AspectOPTIMA.abortTransaction();
	}
	
	/**
	 * Aborts the transaction that is associated to the currently active thread.
	 * @return true if the current transaction was successfully aborted, false in case of any error
	 */
	public static boolean abortTransactionOfForeignThread(OMTTThread thread) {
		try {
			OMTTParticipant participant = getParticipantFromThread(thread);
			participant.abortTransaction();
			return true;
		} catch (Exception e) {
			/* swallow */
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Sets the minimal number of participants that is required to begin a transaction
	 * to the passed value for the transaction of the current participant.
	 */
	public static void setMinParticipantCountInCurrentThread(int min) {
		setMinParticipantCount(min);
	}
	
	public static void setMinParticipantCountInExistingThread(OMTTThread thread, final int min) {
		Runnable setMinParticipantCountRunnable = new Runnable() {
					public void run() {
						setMinParticipantCount(min);
					}
		};
		thread.enqueue((Runnable) setMinParticipantCountRunnable);
	}
	
	private static void setMinParticipantCount(int min) {
		OMTTParticipant participant = (OMTTParticipant) ParticipantPool.claimParticipant();
		mapKeyToThread(participant, (OMTTThread) Thread.currentThread());
		OMTTContext context = (OMTTContext) participant.getContext();
		context.setMinParticipantCount(min);
	}
	
	/**
	 * Maps the passed key object to the passed thread for futher reference. 
	 */
	public static <T> void mapKeyToThread(T key, OMTTThread thread) {
		keyToThreadMap.put(key, thread);
	}
	
	/**
	 * Returns the thread that is mapped to the passed key.
	 */
	public static <T> OMTTThread getThread(T key) {
		return keyToThreadMap.get(key);
	}
}